/* 
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package java.util.logging;

import java.util.Date;
import java.util.ResourceBundle;

/**
 * Format a given <code>LogRecord</code> into string represents XML. The DTD specified in 
 * Appendix A to Java Logging APIs specification is used.
 * 
 * <code>XMLFormatter</code> uses given <code>Handler</code>'s encoding if has, otherwise
 * uses default platform encoding instead. However, the UTF-8 is recommended encoding. 
 * 
 */
public class XMLFormatter extends Formatter {

    private static final String lineSeperator = "/";

    private static final String indent = "    "; //$NON-NLS-1$

    /**
     * Default constructor
     */
    public XMLFormatter() {
        super();
    }

    /**
     * Format a <code>LogRecord</code> into string which represents XML 
     * 
     * @param r the given LogRecord instance to be formatted
     * @return string which represents XML 
     */
    public String format(LogRecord r) {
        //call a method of LogRecord to ensure not null
        long time = r.getMillis();
        //format to date
        String date = new Date(time).toString();

        StringBuffer sb = new StringBuffer();
        sb.append(("<record>")).append(lineSeperator); //$NON-NLS-1$
        sb.append(indent).append(("<date>")).append(date).append(("</date>")) //$NON-NLS-1$ //$NON-NLS-2$
                .append(lineSeperator);
        sb.append(indent).append(("<millis>")).append(time).append( //$NON-NLS-1$
                ("</millis>")).append(lineSeperator); //$NON-NLS-1$
        sb.append(indent).append(("<sequence>")).append(r.getSequenceNumber()) //$NON-NLS-1$
                .append(("</sequence>")).append(lineSeperator); //$NON-NLS-1$
        if (null != r.getLoggerName()) {
            sb.append(indent).append(("<logger>")).append(r.getLoggerName()) //$NON-NLS-1$
                    .append(("</logger>")).append(lineSeperator); //$NON-NLS-1$
        }
        sb.append(indent).append(("<level>")).append(r.getLevel().getName()) //$NON-NLS-1$
                .append(("</level>")).append(lineSeperator); //$NON-NLS-1$
        if (null != r.getSourceClassName()) {
            sb.append(indent).append(("<class>")) //$NON-NLS-1$
                    .append(r.getSourceClassName()).append(("</class>")) //$NON-NLS-1$
                    .append(lineSeperator);
        }
        if (null != r.getSourceMethodName()) {
            sb.append(indent).append(("<method>")).append( //$NON-NLS-1$
                    r.getSourceMethodName()).append(("</method>")).append( //$NON-NLS-1$
                    lineSeperator);
        }
        sb.append(indent).append(("<thread>")).append(r.getThreadID()).append( //$NON-NLS-1$
                ("</thread>")).append(lineSeperator); //$NON-NLS-1$
        formatMessages(r, sb);
        Object[] params;
        if ((params = r.getParameters()) != null) {
            for (int i = 0; i < params.length; i++) {
                Object element = params[i];
                sb.append(indent).append(("<param>")).append(element).append( //$NON-NLS-1$
                        ("</param>")).append(lineSeperator); //$NON-NLS-1$
            }
        }
        formatThrowable(r, sb);
        sb.append(("</record>")).append(lineSeperator); //$NON-NLS-1$
        return sb.toString();
    }

    private void formatMessages(LogRecord r, StringBuffer sb) {
        //get localized message if has, but don't call Formatter.formatMessage to parse pattern string
        ResourceBundle rb = r.getResourceBundle();
        String pattern = r.getMessage();
        if (null != rb && null != pattern) {
            String message;
            try {
                message = rb.getString(pattern);
            } catch (Exception e) {
                message = null;
            }

            if (message == null) {
                message = pattern;
                sb.append(indent).append(("<message>")).append(message).append( //$NON-NLS-1$
                        ("</message>")).append(lineSeperator); //$NON-NLS-1$
            } else {
                sb.append(indent).append(("<message>")).append(message).append( //$NON-NLS-1$
                        ("</message>")).append(lineSeperator); //$NON-NLS-1$
                sb.append(indent).append(("<key>")).append(pattern).append( //$NON-NLS-1$
                        ("</key>")).append(lineSeperator); //$NON-NLS-1$
                sb.append(indent).append(("<catalog>")).append( //$NON-NLS-1$
                        r.getResourceBundleName()).append(("</catalog>")) //$NON-NLS-1$
                        .append(lineSeperator);
            }
        } else if(null != pattern){
            sb.append(indent).append(("<message>")).append(pattern).append( //$NON-NLS-1$
                    ("</message>")).append(lineSeperator); //$NON-NLS-1$
        } else{
            sb.append(indent).append(("<message/>")); //$NON-NLS-1$
        }
    }

    private void formatThrowable(LogRecord r, StringBuffer sb) {
        Throwable t;
        if ((t = r.getThrown()) != null) {
            sb.append(indent).append("<exception>").append(lineSeperator); //$NON-NLS-1$
            sb.append(indent).append(indent).append("<message>").append( //$NON-NLS-1$
                    t.toString()).append("</message>").append(lineSeperator); //$NON-NLS-1$
            //format throwable's stack trace
            StackTraceElement[] elements = t.getStackTrace();
            for (int i = 0; i < elements.length; i++) {
                StackTraceElement e = elements[i];
                sb.append(indent).append(indent).append("<frame>").append( //$NON-NLS-1$
                        lineSeperator);
                sb.append(indent).append(indent).append(indent).append(
                        "<class>").append(e.getClassName()).append("</class>")  //$NON-NLS-1$//$NON-NLS-2$
                        .append(lineSeperator);
                sb.append(indent).append(indent).append(indent).append(
                        "<method>").append(e.getMethodName()).append( //$NON-NLS-1$
                        "</method>").append(lineSeperator); //$NON-NLS-1$
                sb.append(indent).append(indent).append(indent)
                        .append("<line>").append(e.getLineNumber()).append( //$NON-NLS-1$
                        "</line>").append(lineSeperator); //$NON-NLS-1$
                sb.append(indent).append(indent).append("</frame>").append( //$NON-NLS-1$
                        lineSeperator);
            }
            sb.append(indent).append("</exception>").append(lineSeperator); //$NON-NLS-1$
        }
    }

    /**
     * Return the header string for XML, use given handler's encoding if has, 
     * other wise use default platform encoding 
     * 
     * @param h the given handler
     * @return the header string for XML
     */
    public String getHead(Handler h) {
        String encoding = null;
        if(null != h) {
        	encoding = h.getEncoding();
        }
        if (null == encoding) {
            encoding = ""; // FIXME? detect somehow
        }
        StringBuffer sb = new StringBuffer();
        sb.append("<?xml version=\"1.0\" encoding=\"").append(encoding).append( //$NON-NLS-1$
                "\" standalone=\"no\"?>").append(lineSeperator); //$NON-NLS-1$
        sb.append("<!DOCTYPE log SYSTEM \"logger.dtd\">").append(lineSeperator); //$NON-NLS-1$
        sb.append(("<log>")); //$NON-NLS-1$
        return sb.toString();
    }

    /**
     * Return the tail string for XML
     * 
     * @param h the given handler
     * @return the tail string for XML
     */
    public String getTail(Handler h) {
        return "</log>"; //$NON-NLS-1$
    }

}


